import logging

access_log = logging.getLogger("hurricane.server.access")
app_log = logging.getLogger("hurricane.server.application")
logger = logging.getLogger("hurricane.server.general")
metrics_log = logging.getLogger("hurricane.server.metrics")
